pub use super::Game;
use serde::{Serialize, Serializer};
use std::fmt;

macro_rules! draw {
    ($g:ident, $e:expr) => {
        $g.players[$g.active_player].draw($e)
    };
}

macro_rules! action {
    ($g:ident, $e:expr) => {
        $g.turn_state.actions += $e
    };
}

macro_rules! buy {
    ($g:ident, $e:expr) => {
        $g.turn_state.buys += $e
    };
}

macro_rules! coin {
    ($g:ident, $e:expr) => {
        $g.turn_state.coin += $e
    };
}

fn serialize_card_type<S>(_: &fn(&mut super::Game), serializer: S) -> Result<S::Ok, S::Error>
where
    S: Serializer,
{
    serializer.serialize_str("ActionSer")
}

fn serialize_card_type_0<S>(
    _: &fn(&super::Game, usize) -> u32,
    serializer: S,
) -> Result<S::Ok, S::Error>
where
    S: Serializer,
{
    serializer.serialize_str("ActionSer")
}

#[derive(Clone, Serialize)]
pub enum CardType {
    #[serde(serialize_with = "serialize_card_type")]
    Action(fn(&mut Game)),
    Attack,
    Curse,
    #[serde(serialize_with = "serialize_card_type")]
    Reaction(fn(&mut Game)),
    Treasure(u32),
    #[serde(serialize_with = "serialize_card_type_0")]
    Victory(fn(&Game, usize) -> u32),
}

#[derive(Clone)]
pub struct Card {
    pub cost: u32,
    pub name: &'static str,
    pub types: Vec<CardType>,
}

impl Card {
    pub fn new(name: &'static str, cost: u32) -> Card {
        Card {
            name,
            cost,
            types: vec![],
        }
    }

    pub fn with_type(&self, card_type: CardType) -> Card {
        let mut card = self.clone();
        card.types.push(card_type);
        card
    }

    pub fn action(&self) -> Option<fn(&mut Game)> {
        for t in &self.types {
            match t {
                CardType::Action(effects) => return Some(*effects),
                _ => (),
            }
        }
        None
    }

    pub fn curse(&self) -> Option<()> {
        for t in &self.types {
            match t {
                CardType::Curse => return Some(()),
                _ => (),
            }
        }
        None
    }

    pub fn treasure(&self) -> Option<u32> {
        for t in &self.types {
            match t {
                CardType::Treasure(coin) => return Some(*coin),
                _ => (),
            }
        }
        None
    }

    pub fn victory(&self) -> Option<fn(&Game, usize) -> u32> {
        for t in &self.types {
            match t {
                CardType::Victory(points) => return Some(*points),
                _ => (),
            }
        }
        None
    }
}

impl Serialize for Card {
    fn serialize<S>(&self, serializer: S) -> Result<S::Ok, S::Error>
    where
        S: Serializer,
    {
        serializer.serialize_str(&self.name)
    }
}

impl fmt::Display for Card {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        write!(f, "{}", self.name)
    }
}
