#[macro_use]
mod card;
mod sets;

use async_std::{prelude::*, sync::RwLock};
use card::*;
use itertools::Itertools;
use rand::{seq::SliceRandom, thread_rng, Rng};
use serde::{Deserialize, Serialize};
use std::{collections::HashMap, sync::Arc};
use tide::{Body, Redirect, Request, Response};
use tide_websockets::{Message, WebSocket, WebSocketConnection};
use uuid::Uuid;

#[derive(Deserialize)]
#[serde(tag = "type")]
enum ClientMessage {
    Chat { message: String },
    CreateGame { name: String },
    JoinGame { name: String },
    Command { command: Command },
}

#[derive(Serialize)]
#[serde(tag = "type")]
enum ServerMessage {
    Chat {
        sender: String,
        message: String,
    },
    PlayerJoined {
        player: String,
    },
    GameState {
        state: GameState,
        players: Vec<PlayerState>,
        active_player: usize,
        supply: Vec<PileState>,
        turn_state: TurnState,
        trash: Vec<Card>,
    },
    GameSetup {
        setup: GameSetup,
    },
    PlayerHand {
        hand: Vec<Card>,
    },
    PlayerId {
        id: usize,
    },
    Notification {
        event: Event,
    },
    GameOver {
        score: Vec<(usize, u32)>,
    },
    ResolveRequest {
        /// Request type and context
        request: ResolveRequest,
        /// The card that is the source of the request to display to the player
        card: String,
    },
}

#[derive(Clone, Serialize)]
struct GameSetup {
    supply: Vec<Card>,
    deck: Vec<Card>,
}

#[derive(Serialize)]
struct PlayerState {
    id: usize,
    name: String,
    draw_pile_count: usize,
    hand_count: usize,
    discard_pile: Option<String>,
    played_cards: Vec<String>,
}

#[derive(Serialize)]
struct PileState {
    name: Card,
    count: usize,
}

#[derive(Clone, Serialize)]
struct TurnState {
    actions: u32,
    buys: u32,
    coin: u32,
}

impl Default for TurnState {
    fn default() -> Self {
        TurnState {
            actions: 1,
            buys: 1,
            coin: 0,
        }
    }
}

struct Player {
    id: String,
    name: String,
    draw_pile: Vec<Card>,
    hand: Vec<Card>,
    discard_pile: Vec<Card>,
    played_cards: Vec<Card>,
}

impl Player {
    pub fn new(id: String, name: String) -> Self {
        Self {
            id,
            name,
            draw_pile: vec![],
            hand: vec![],
            discard_pile: vec![],
            played_cards: vec![],
        }
    }

    pub fn draw(&mut self, count: usize) {
        for _ in 0..count {
            if self.draw_pile.len() == 0 {
                self.draw_pile.append(&mut self.discard_pile);
                self.draw_pile.shuffle(&mut thread_rng());
            }

            if self.draw_pile.len() > 0 {
                self.hand.push(self.draw_pile.pop().unwrap());
            }
        }
    }

    pub fn discard(&mut self, index: usize) {
        self.discard_pile.push(self.hand.remove(index));
    }
}

#[derive(Copy, Clone, PartialEq, Serialize)]
enum GameState {
    Setup,
    InProgress,
    Over,
}

impl Default for GameSetup {
    fn default() -> GameSetup {
        let set = sets::base::base();

        GameSetup {
            supply: vec![
                set.get("Chapel").unwrap(),
                set.get("Workshop").unwrap(),
                set.get("Bureaucrat").unwrap(),
                set.get("Gardens").unwrap(),
                set.get("Throne Room").unwrap(),
                set.get("Bandit").unwrap(),
                set.get("Festival").unwrap(),
                set.get("Sentry").unwrap(),
                set.get("Witch").unwrap(),
                set.get("Artisan").unwrap(),
            ],
            deck: vec![
                set.get("Copper").unwrap(),
                set.get("Copper").unwrap(),
                set.get("Copper").unwrap(),
                set.get("Copper").unwrap(),
                set.get("Copper").unwrap(),
                set.get("Copper").unwrap(),
                set.get("Copper").unwrap(),
                set.get("Estate").unwrap(),
                set.get("Estate").unwrap(),
                set.get("Estate").unwrap(),
            ],
        }
    }
}

/// Which player's input is requested to resolve the current effect
#[derive(Clone, Serialize, PartialEq)]
enum ResolvingPlayer {
    ActivePlayer,
    AllNonActivePlayers,
}

#[derive(Deserialize)]
#[serde(tag = "type")]
enum ResolveReply {
    HandCardsChosen { choice: Vec<usize> },
    SupplyCardChosen { choice: usize },
}

/// Sent by the server to tell the client what information is requested
/// from the player to resolve the current effect
#[derive(Clone, Serialize)]
#[serde(tag = "type")]
enum ResolveRequest {
    ChooseHandCardsToDiscard { filter: CardFilter },
    GainCard { filter: CardFilter },
}

enum ResolveResult {
    Resolved,
    NotResolved,
}

#[derive(Clone)]
enum Effect {
    /// Trigger effect if another card is played while effect is active
    /// return true to indicate the event is resolved and should be removed
    OnCardPlayed(fn(&mut Game, &Card) -> bool),
    /// Effect that blocks further processing of the game state until
    /// some user input is received that allows to fully resolve the effect
    Resolving {
        card: String,
        player: ResolvingPlayer,
        request: ResolveRequest,
        effect: ResolvingEffectHandler,
    },
}

#[derive(Default)]
struct EffectState {
    resolved: bool,
    players_responded: Vec<usize>,
}

#[derive(Clone, Serialize)]
//#[serde(tag = "type")]
enum CardFilter {
    Any,
    And(Vec<CardFilter>),
    MaxCost(u32),
    Type(CardType),
}

impl CardFilter {
    fn match_card(&self, card: &Card) -> bool {
        match &self {
            CardFilter::Any => true,
            CardFilter::And(filter) => filter.iter().all(|f| f.match_card(&card)),
            CardFilter::MaxCost(cost) => card.cost <= *cost,
            CardFilter::Type(card_type) => match card_type {
                CardType::Action(_) => card.action().is_some(),
                CardType::Attack => {
                    unimplemented!()
                }
                CardType::Curse => {
                    unimplemented!()
                }
                CardType::Reaction(_) => {
                    unimplemented!()
                }
                CardType::Treasure(_) => card.treasure().is_some(),
                CardType::Victory(_) => {
                    unimplemented!()
                }
            },
        }
    }
}

type ResolvingEffectHandler =
    fn(&mut Game, &ResolveReply, usize, &ResolveRequest, &mut EffectState) -> ResolveResult;

#[derive(Deserialize)]
#[serde(tag = "type")]
enum Command {
    BuyCard {
        index: usize,
    },
    Discard {
        index: usize,
    },
    DrawCard,
    EndTurn,
    GainCard {
        index: usize,
    },
    PlayCard {
        index: usize,
    },
    ResolveReply {
        reply: ResolveReply,
    },
    StartGame,
    TrashHand {
        index: usize,
    },

    /// Replace a kingdom card in the supply by another one,
    /// only viable during Setup state
    ChangeSupply {
        index: usize,
        name: String,
    },
}

#[derive(Debug, Serialize)]
#[serde(tag = "type")]
enum Event {
    CardBought { player: usize, name: &'static str },
    CardDrawn { player: usize },
    CardGained { player: usize, index: usize },
    CardDiscarded { player: usize },
    CardTrashed { player: usize, name: &'static str },
    CardPlayed { player: usize, name: &'static str },
    CardRevealed { player: usize, name: &'static str },
    TurnStarted { player: usize },
}

pub struct Game {
    effects: Vec<Effect>,
    players: Vec<Player>,
    state: GameState,
    setup: GameSetup,
    connections: HashMap<Uuid, (String, WebSocketConnection)>,
    active_player: usize,
    supply: Vec<(Card, usize)>,
    trash: Vec<Card>,
    turn_state: TurnState,
    debug_mode: bool,

    /// Any effect from a card that requires further input from players
    /// and blocks the game until fully resolved.
    resolving_effect: Option<(
        String,
        ResolveRequest,
        ResolvingEffectHandler,
        ResolvingPlayer,
        EffectState,
    )>,
}

impl Game {
    fn new(player: Player) -> Self {
        Self {
            effects: vec![],
            players: vec![player],
            state: GameState::Setup,
            setup: GameSetup::default(),
            connections: HashMap::new(),
            active_player: 0,
            supply: vec![],
            trash: vec![],
            turn_state: TurnState::default(),
            resolving_effect: None,
            debug_mode: true,
        }
    }

    fn supply_name_to_index(&self, name: &String) -> Option<usize> {
        for i in 0..self.supply.len() {
            if self.supply[i].0.name == name {
                return Some(i);
            }
        }
        None
    }

    fn start(&mut self) {
        match self.state {
            GameState::Setup => {
                self.state = GameState::InProgress;

                if self.players.len() > 1 {
                    self.active_player = thread_rng().gen_range(0..self.players.len());
                }

                for p in self.players.iter_mut() {
                    (*p).discard_pile = self.setup.deck.clone();
                    p.draw(5);
                }

                let victory_qty = match self.players.len() {
                    x if x <= 2 => 8,
                    _ => 12,
                };

                self.supply = vec![
                    (
                        sets::base::base().get("Copper").unwrap(),
                        60 - self.players.len() * 7,
                    ),
                    (sets::base::base().get("Silver").unwrap(), 40),
                    (sets::base::base().get("Gold").unwrap(), 30),
                    (sets::base::base().get("Estate").unwrap(), victory_qty),
                    (sets::base::base().get("Duchy").unwrap(), victory_qty),
                    (sets::base::base().get("Province").unwrap(), victory_qty),
                    (sets::base::base().get("Curse").unwrap(), 10),
                ];

                self.setup
                    .supply
                    .sort_by(|a, b| a.cost.cmp(&b.cost).then(a.name.cmp(&b.name)));
                for supply_card in self.setup.supply.iter() {
                    self.supply.push((supply_card.clone(), 10));
                }
            }

            _ => {} // Ignore if game is not in setup state
        }
    }

    fn end_turn(&mut self) {
        if let Some(_) = self.resolving_effect {
            return;
        }

        match self.players.get_mut(self.active_player) {
            None => {}
            Some(p) => {
                p.discard_pile.append(&mut p.played_cards);
                p.discard_pile.append(&mut p.hand);
                p.draw(5);
            }
        }

        self.end_game();

        self.active_player += 1;
        self.active_player %= self.players.len();
        self.turn_state = TurnState::default();

        self.effects.clear();

        self.emit(Event::TurnStarted {
            player: self.active_player,
        });
    }

    // Check if the end game condition is reached and finish the game if so,
    // sending a message to all clients.
    fn end_game(&mut self) {
        let provinces = self
            .supply
            .iter()
            .any(|(c, n)| (*c).name == "Province" && *n == 0);
        let supply = self.supply.iter().filter(|(_, n)| *n == 0).count() > 2;

        if supply || provinces {
            self.state = GameState::Over;

            for p in self.players.iter_mut() {
                p.draw_pile.append(&mut p.played_cards);
                p.draw_pile.append(&mut p.hand);
                p.draw_pile.append(&mut p.discard_pile);
            }
        }
    }

    fn is_active_player(&self, player: usize) -> bool {
        self.active_player == player
    }

    fn get_active_player(&mut self) -> &mut Player {
        return &mut self.players[self.active_player];
    }

    pub fn trash_hand(&mut self, player_number: usize, index: usize) {
        if let Some(_) = self.resolving_effect {
            return;
        }

        self.trash
            .push(self.players[player_number].hand.remove(index));
    }

    pub fn can_play(&self, player_number: usize, index: usize) -> bool {
        if let Some(_) = self.resolving_effect {
            return false;
        }

        let card = self.players[player_number].hand.get(index).unwrap();

        match card.action() {
            Some(_) => {
                if self.turn_state.actions > 0 {
                    return true;
                }
            }
            None => (),
        }

        match card.treasure() {
            Some(_) => {
                return true;
            }
            None => (),
        }

        false
    }

    pub fn play_card(&mut self, player_number: usize, index: usize) -> bool {
        if !self.can_play(player_number, index) {
            return false;
        }

        let card = self.players[player_number].hand.remove(index);
        self.emit(Event::CardPlayed {
            player: player_number,
            name: card.name,
        });

        if let Some(coin) = card.treasure() {
            self.turn_state.coin += coin;
        }

        if let Some(effect) = card.action() {
            self.turn_state.actions -= 1;

            effect(self);
        }

        let mut effects = self.effects.clone();
        self.effects.clear();
        effects.retain(|effect| {
            match effect {
                Effect::OnCardPlayed(effect) => {
                    if effect(self, &card) {
                        return false;
                    }
                }

                Effect::Resolving { .. } => {}
            }

            true
        });
        self.effects.append(&mut effects);

        self.players[player_number].played_cards.push(card);
        true
    }

    fn gain_card(&mut self, player: usize, index: &usize, filter: &CardFilter) -> Result<(), ()> {
        if let Some((card, count)) = self.supply.get(*index) {
            if *count < 1 {
                return Err(());
            }

            if !filter.match_card(&card) {
                return Err(());
            }

            self.supply.get_mut(*index).unwrap().1 = self.supply.get(*index).unwrap().1 - 1;
            let card = self.supply[*index].0.clone();

            self.players[player].discard_pile.push(card.clone());

            self.emit(Event::CardGained {
                player,
                index: *index,
            });

            return Ok(());
        }
        Err(())
    }

    pub fn buy_card(&mut self, player_number: usize, index: usize) -> bool /*-> Result<(), &'static str>*/
    {
        if let Some(_) = self.resolving_effect {
            return false;
        }

        if player_number != self.active_player {
            return false;
        }

        if self.turn_state.buys < 1 {
            return false;
        }

        if let Some(card) = self.supply.get(index).as_deref() {
            if card.0.cost <= self.turn_state.coin {
                let card = card.0.clone();

                self.supply.get_mut(index).unwrap().1 = self.supply.get(index).unwrap().1 - 1;
                self.turn_state.coin -= card.cost;
                self.turn_state.buys -= 1;
                self.players[player_number].discard_pile.push(card);
                return true;
                //return Ok(());
            }
        }
        //Err("Not enough coin");
        false
    }

    fn add_effect(&mut self, effect: Effect) {
        match effect {
            Effect::OnCardPlayed(_) => {
                self.effects.push(effect);
            }

            Effect::Resolving {
                card,
                request,
                effect,
                player,
            } => {
                if player == ResolvingPlayer::AllNonActivePlayers && self.players.len() == 1 {
                    return;
                }

                let state = EffectState::default();
                self.resolving_effect = Some((card, request, effect, player, state));
            }
        }
    }

    fn handle_command(&mut self, player: usize, command: Command) {
        match command {
            Command::ChangeSupply { index, name } => {
                if let GameState::Setup = self.state {
                    if let Some(card) = sets::base::base().get(name.as_ref()) {
                        self.setup.supply[index] = card;
                    }
                }
            }

            Command::DrawCard => {
                if !self.debug_mode {
                    return;
                }

                self.players[player].draw(1);
                self.emit(Event::CardDrawn { player });
            }

            Command::BuyCard { index } => {
                if self.buy_card(player, index) {
                    self.emit(Event::CardBought {
                        player,
                        name: self.supply[index].0.name,
                    });
                }
            }

            Command::EndTurn => {
                if self.is_active_player(player) {
                    self.end_turn();
                }
            }

            Command::GainCard { index } => {
                if !self.debug_mode {
                    return;
                }

                self.gain_card(player, &index, &CardFilter::Any);
            }

            Command::Discard { index } => {
                if !self.debug_mode {
                    return;
                }

                //let card_name = self.players[player].hand[index].clone();
                self.players[player].discard(index);
                self.emit(Event::CardDiscarded { player });
            }

            Command::StartGame => {
                self.start();
            }

            Command::TrashHand { index } => {
                let name = self.players[player].hand[index].name.clone();
                self.trash_hand(player, index);
                self.emit(Event::CardTrashed { player, name });
            }

            Command::PlayCard { index } => {
                self.play_card(player, index);
            }

            Command::ResolveReply { reply } => {
                if let Some((card, request, effect, resolve_player, mut state)) =
                    self.resolving_effect.take()
                {
                    match resolve_player {
                        ResolvingPlayer::ActivePlayer => {
                            if player == self.active_player {
                                match effect(self, &reply, self.active_player, &request, &mut state)
                                {
                                    ResolveResult::Resolved => state.resolved = true,
                                    ResolveResult::NotResolved => (),
                                }
                            }
                        }

                        ResolvingPlayer::AllNonActivePlayers => {
                            if player != self.active_player {
                                match effect(self, &reply, player, &request, &mut state) {
                                    ResolveResult::Resolved => {
                                        state.players_responded.push(player);
                                    }
                                    ResolveResult::NotResolved => (),
                                }
                            }

                            if state.players_responded.len() == self.players.len() - 1 {
                                state.resolved = true;
                            }
                        }
                    }

                    match state.resolved {
                        false => {
                            self.resolving_effect =
                                Some((card, request, effect, resolve_player, state));
                        }
                        true => {}
                    }
                }
            }
        }
    }

    fn emit(&self, event: Event) {
        println!("Emitting event: {:?}", event);
        async_std::task::block_on(notify_players(self, event));
    }
}

#[derive(Clone, Default)]
struct State {
    games: Arc<RwLock<HashMap<Uuid, Game>>>,
}

async fn notify_players(game: &Game, event: Event) {
    let sm = ServerMessage::Notification { event };
    broadcast(game, &sm).await;
}

async fn broadcast(game: &Game, sm: &ServerMessage) {
    for (_, (_, con)) in game.connections.iter() {
        con.send_json(&sm).await;
    }
}

async fn broadcast_state(game: &Game) {
    let sm = ServerMessage::GameState {
        state: game.state,
        players: game
            .players
            .iter()
            .enumerate()
            .map(|(i, p)| PlayerState {
                id: i,
                name: html_escape::encode_text(&p.name).into(),
                draw_pile_count: p.draw_pile.len(),
                hand_count: p.hand.len(),
                discard_pile: p.discard_pile.last().map(|c| c.name.clone().into()),
                played_cards: p
                    .played_cards
                    .iter()
                    .map(|c| c.name.clone().into())
                    .collect(),
            })
            .collect(),
        active_player: game.active_player,
        supply: game
            .supply
            .iter()
            .map(|(card, count)| PileState {
                name: card.clone(),
                count: count.clone(),
            })
            .collect(),
        turn_state: game.turn_state.clone(),
        trash: game.trash.clone(),
    };

    broadcast(&game, &sm).await;

    if let GameState::Setup = game.state {
        let sm = ServerMessage::GameSetup {
            setup: game.setup.clone(),
        };
        broadcast(&game, &sm).await;
    }

    for p in game.players.iter() {
        let sm = ServerMessage::PlayerHand {
            hand: p.hand.clone(),
        };

        send_msg(&game, &p, &sm).await;
    }

    if game.state == GameState::Over {
        let sm = ServerMessage::GameOver {
            score: game
                .players
                .iter()
                .enumerate()
                .map(|(i, p)| {
                    let score = p.draw_pile.iter().fold(0, |acc, card| {
                        if let Some(points) = card.victory() {
                            acc + points(&game, i)
                        } else if let Some(_) = card.curse() {
                            acc - 1
                        } else {
                            acc
                        }
                    });
                    (i, score)
                })
                .sorted_by(|a, b| Ord::cmp(&b.1, &a.1))
                .collect::<Vec<_>>(),
        };
        broadcast(&game, &sm).await;
    }

    if let Some((card_name, request, _, ref player, _)) = &game.resolving_effect {
        match request {
            ResolveRequest::ChooseHandCardsToDiscard { .. } => match player {
                ResolvingPlayer::ActivePlayer => {
                    let p = game.players.get(game.active_player).unwrap();
                    let sm = ServerMessage::ResolveRequest {
                        card: card_name.clone(),
                        request: request.clone(),
                    };
                    send_msg(&game, &p, &sm).await;
                }

                ResolvingPlayer::AllNonActivePlayers => {
                    for (id, player) in game.players.iter().enumerate() {
                        let sm = ServerMessage::ResolveRequest {
                            card: card_name.clone(),
                            request: request.clone(),
                        };

                        if id != game.active_player {
                            send_msg(&game, &player, &sm).await;
                        }
                    }
                }
            },

            ResolveRequest::GainCard { .. } => match player {
                ResolvingPlayer::ActivePlayer => {
                    let p = game.players.get(game.active_player).unwrap();
                    let sm = ServerMessage::ResolveRequest {
                        card: card_name.clone(),
                        request: request.clone(),
                    };
                    send_msg(&game, &p, &sm).await;
                }

                _ => unimplemented!("GainCard for non active players?!"),
            },
        }
    }
}

async fn send_msg(game: &Game, player: &Player, sm: &ServerMessage) {
    for (_, (player_id, con)) in game.connections.iter() {
        if *player_id == (*player).id {
            con.send_json(&sm).await.unwrap();
        }
    }
}

#[async_std::main]
async fn main() -> Result<(), std::io::Error> {
    tide::log::with_level(tide::log::LevelFilter::Debug);

    let mut app = tide::with_state(State::default());

    app.with(tide::sessions::SessionMiddleware::new(
        tide::sessions::MemoryStore::new(),
        "1234567890abcdef1234567890abcdef".as_bytes(),
    ));

    app.at("/")
        .get(|_| async { Ok(Body::from_file("static/index.html").await?) });

    app.at("/static").serve_dir("static/")?;

    app.at("/game").post(|mut req: Request<State>| async move {
        let msg: ClientMessage = req.body_form().await?;
        let session = req.session_mut();

        match msg {
            ClientMessage::CreateGame { name } => {
                session.insert("player_name", name.clone())?;
                let player_id: String = session.id().clone().into();

                let id = Uuid::new_v4();
                let mut games = req.state().games.write().await;
                let player = Player::new(player_id, name);

                games.insert(id, Game::new(player));

                let url = format!("/game/{}", id.to_simple().to_string());
                let res: tide::Result = Ok(Redirect::new(url).into());
                res
            }
            _ => Ok(Response::new(400)),
        }
    });

    app.at("/game/:id").get(|req: Request<State>| async move {
        let games = req.state().games.read().await;
        let id = req.param("id")?;
        let game_id = Uuid::parse_str(&id).unwrap();
        let session = req.session();

        match games.get(&game_id) {
            None => return Ok(Response::new(404)),
            Some(_) => (),
        };

        match session.get::<String>("player_name") {
            None => {
                let url = format!("/game/{}/join", id);
                return Ok(Redirect::new(url).into());
            }
            Some(_) => (),
        };

        let mut res = Response::new(200);
        res.set_body(Body::from_file("static/game.html").await?);
        Ok(res)
    });

    app.at("/game/:id/join")
        .get(|_| async { Ok(Body::from_file("static/join.html").await?) })
        .post(|mut req: Request<State>| async move {
            let msg: ClientMessage = req.body_form().await?;
            let id: String = { req.param("id").unwrap().into() };
            let session = req.session_mut();

            match msg {
                ClientMessage::JoinGame { name } => {
                    session.insert("player_name", name.clone())?;

                    let url = format!("/game/{}", id);
                    let res: tide::Result = Ok(Redirect::new(url).into());
                    res
                }
                _ => Ok(Response::new(500)),
            }
        });

    app.at("/game/:id/ws").get(WebSocket::new(
        |req: Request<State>, mut stream| async move {
            let id = req.param("id")?;
            let game_id = Uuid::parse_str(&id).unwrap();
            let mut games = req.state().games.write().await;
            let game = games.get_mut(&game_id).unwrap();
            let session = req.session();

            let player_id: String = session.id().clone().into();

            let client_id = Uuid::new_v4();
            game.connections
                .insert(client_id, (player_id.clone(), stream.clone()));

            let player_name: String = match session.get("player_name") {
                Some(p) => p,
                None => return Ok(()),
            };

            if game.state == GameState::Setup && !game.players.iter().any(|p| p.id == player_id) {
                game.players
                    .push(Player::new(player_id.clone(), player_name.clone()));
            }

            let player_number = game
                .players
                .iter()
                .enumerate()
                .filter(|(_, p)| p.id == player_id)
                .fold(0, |_, (i, _)| i);
            let msg = ServerMessage::PlayerId { id: player_number };
            stream.send_json(&msg).await?;

            //Ensure the write locks are freed, possibly better to move to a function
            //with implicit drop?
            drop(game);
            drop(games);

            let games = req.state().games.read().await;
            let game = games.get(&game_id).unwrap();

            broadcast_state(&game).await;

            let msg = ServerMessage::GameSetup {
                setup: game.setup.clone(),
            };
            stream.send_json(&msg).await?;

            let sm = ServerMessage::PlayerJoined {
                player: html_escape::encode_text(&player_name).into(),
            };

            broadcast(&game, &sm).await;

            drop(game);
            drop(games);

            while let Some(Ok(Message::Text(input))) = stream.next().await {
                let msg: ClientMessage = serde_json::from_str(&input)?;

                match msg {
                    ClientMessage::Chat { message } => {
                        let sender: String = session.get("player_name").unwrap();
                        let sm = ServerMessage::Chat {
                            sender: html_escape::encode_text(&sender).into(),
                            message: html_escape::encode_text(&message).into(),
                        };

                        let games = req.state().games.read().await;
                        let game = games.get(&game_id).unwrap();

                        broadcast(&game, &sm).await;
                    }

                    ClientMessage::CreateGame { .. } => {}

                    ClientMessage::JoinGame { .. } => {}

                    ClientMessage::Command { command } => {
                        let mut games = req.state().games.write().await;
                        let game = games.get_mut(&game_id).unwrap();

                        game.handle_command(player_number, command);

                        broadcast_state(&game).await;
                    }
                }
            }

            let mut games = req.state().games.write().await;
            let game = games.get_mut(&game_id).unwrap();

            game.connections.remove(&client_id);

            Ok(())
        },
    ));

    app.listen("0.0.0.0:5000").await?;
    Ok(())
}
