use crate::card::{Card, CardType};
use crate::{
    CardFilter, Effect, Event, ResolveReply, ResolveRequest, ResolveResult::*, ResolvingPlayer,
};
use itertools::Itertools;

pub struct CardSet {
    name: &'static str,
    cards: Vec<Card>,
}

impl CardSet {
    pub fn get(&self, name: &str) -> Option<Card> {
        for card in self.cards.iter() {
            if card.name == name {
                return Some(card.clone());
            }
        }
        None
    }
}

pub fn base() -> CardSet {
    CardSet {
        name: "Dominion",
        cards: vec![
            copper(),
            silver(),
            gold(),
            estate(),
            duchy(),
            province(),
            curse(),
            artisan(),
            bandit(),
            bureaucrat(),
            chapel(),
            cellar(),
            gardens(),
            festival(),
            market(),
            militia(),
            mine(),
            merchant(),
            moat(),
            remodel(),
            sentry(),
            smithy(),
            throne_room(),
            village(),
            witch(),
            workshop(),
        ],
    }
}

fn copper() -> Card {
    Card::new("Copper", 0).with_type(CardType::Treasure(1))
}

fn silver() -> Card {
    Card::new("Silver", 3).with_type(CardType::Treasure(2))
}

fn gold() -> Card {
    Card::new("Gold", 6).with_type(CardType::Treasure(3))
}

fn estate() -> Card {
    Card::new("Estate", 2).with_type(CardType::Victory(|_, _| 1))
}

fn duchy() -> Card {
    Card::new("Duchy", 5).with_type(CardType::Victory(|_, _| 3))
}

fn province() -> Card {
    Card::new("Province", 8).with_type(CardType::Victory(|_, _| 6))
}

fn curse() -> Card {
    Card::new("Curse", 8).with_type(CardType::Curse)
}

fn artisan() -> Card {
    Card::new("Artisan", 6).with_type(CardType::Action(|_game| {}))
}

fn bandit() -> Card {
    Card::new("Bandit", 5)
        .with_type(CardType::Attack)
        .with_type(CardType::Action(|_game| {}))
}

fn bureaucrat() -> Card {
    Card::new("Bureaucrat", 4)
        .with_type(CardType::Attack)
        .with_type(CardType::Action(|game| {
            if let Some(index) = game.supply_name_to_index(&"Silver".into()) {
                game.gain_card(game.active_player, &index, &CardFilter::Any);
                let card = game.players[game.active_player].discard_pile.pop().unwrap();
                game.players[game.active_player].draw_pile.push(card);
            }

            game.add_effect(Effect::Resolving {
                card: "Bureaucrat".into(),
                request: ResolveRequest::ChooseHandCardsToDiscard {
                    filter: CardFilter::Type(CardType::Victory(|_, _| 0)), //FIXME!
                },
                player: ResolvingPlayer::AllNonActivePlayers,
                effect: |game, message, player, _request, _state| {
                    if let ResolveReply::HandCardsChosen { choice } = message {
                        if choice.len() > 1 {
                            return NotResolved;
                        }

                        if choice.len() == 0 {
                            if game.players[player].hand.iter().any(|c| c.name == "Moat") {
                                game.emit(Event::CardRevealed {
                                    player,
                                    name: "Moat",
                                });
                                return Resolved;
                            }

                            if !game.players[player]
                                .hand
                                .iter()
                                .any(|c| c.victory().is_some())
                            {
                                for card in game.players[player].hand.iter() {
                                    game.emit(Event::CardRevealed {
                                        player,
                                        name: card.name,
                                    });
                                }
                                return Resolved;
                            }
                        }

                        if let Some(card) = game.players[player].hand.get(choice[0]) {
                            if card.victory().is_none() {
                                return NotResolved;
                            }
                        }

                        let card = game.players[player].hand.remove(choice[0]);
                        game.emit(Event::CardRevealed {
                            player,
                            name: card.name,
                        });
                        game.players[player].draw_pile.push(card);

                        return Resolved;
                    }

                    NotResolved
                },
            });
        }))
}

fn cellar() -> Card {
    Card::new("Cellar", 2).with_type(CardType::Action(|game| {
        action!(game, 1);
        game.add_effect(Effect::Resolving {
            card: "Cellar".into(),
            request: ResolveRequest::ChooseHandCardsToDiscard {
                filter: CardFilter::Any,
            },
            player: ResolvingPlayer::ActivePlayer,
            effect: |game, message, _player, _request, _state| match message {
                ResolveReply::HandCardsChosen { choice } => {
                    let mut discarded = 0;
                    for c in choice.iter().sorted_by(|a, b| b.cmp(a)) {
                        game.get_active_player().discard(*c);
                        discarded += 1;
                    }

                    game.players[game.active_player].draw(discarded);
                    Resolved
                }
                _ => NotResolved,
            },
        });
    }))
}

fn chapel() -> Card {
    Card::new("Chapel", 2).with_type(CardType::Action(|game| {
        game.add_effect(Effect::Resolving {
            card: "Chapel".into(),
            request: ResolveRequest::ChooseHandCardsToDiscard {
                filter: CardFilter::Any,
            },
            player: ResolvingPlayer::ActivePlayer,
            effect: |game, message, _player, _request, _state| {
                if let ResolveReply::HandCardsChosen { choice } = message {
                    if choice.len() > 4 {
                        return NotResolved;
                    }

                    for c in choice.iter().sorted_by(|a, b| b.cmp(a)) {
                        let card = game.players[game.active_player].hand.remove(*c);
                        game.trash.push(card);
                    }

                    return Resolved;
                }
                NotResolved
            },
        })
    }))
}

fn festival() -> Card {
    Card::new("Festival", 5).with_type(CardType::Action(|game| {
        action!(game, 2);
        buy!(game, 1);
        coin!(game, 2);
    }))
}

fn gardens() -> Card {
    Card::new("Gardens", 4).with_type(CardType::Victory(|game, player| {
        game.players[player].draw_pile.len() as u32 / 10
    }))
}

fn moat() -> Card {
    Card::new("Moat", 2)
        .with_type(CardType::Action(|game| {
            draw!(game, 2);
        }))
        .with_type(CardType::Reaction(|_| {}))
}

fn sentry() -> Card {
    Card::new("Sentry", 5).with_type(CardType::Action(|game| {
        draw!(game, 1);
        action!(game, 1);
    }))
}

fn throne_room() -> Card {
    Card::new("Throne Room", 4).with_type(CardType::Action(|_game| {}))
}

fn village() -> Card {
    Card::new("Village", 3).with_type(CardType::Action(|game| {
        draw!(game, 1);
        action!(game, 2);
    }))
}

fn merchant() -> Card {
    Card::new("Merchant", 3).with_type(CardType::Action(|game| {
        draw!(game, 1);
        action!(game, 1);
        game.add_effect(Effect::OnCardPlayed(|game, card| {
            if card.name == "Silver" {
                coin!(game, 1);
                true
            } else {
                false
            }
        }));
    }))
}

fn witch() -> Card {
    Card::new("Witch", 5)
        .with_type(CardType::Attack)
        .with_type(CardType::Action(|game| {
            draw!(game, 2);
        }))
}

fn workshop() -> Card {
    Card::new("Workshop", 3).with_type(CardType::Action(|game| {
        game.add_effect(Effect::Resolving {
            card: "Workshop".into(),
            request: ResolveRequest::GainCard {
                filter: CardFilter::MaxCost(4),
            },
            player: ResolvingPlayer::ActivePlayer,
            effect: |game, message, player, request, _state| {
                if let ResolveReply::SupplyCardChosen { choice } = message {
                    if let ResolveRequest::GainCard { filter } = request {
                        if let Ok(()) = game.gain_card(player, choice, filter) {
                            return Resolved;
                        }
                    }
                }
                NotResolved
            },
        });
    }))
}

fn smithy() -> Card {
    Card::new("Smithy", 4).with_type(CardType::Action(|game| draw!(game, 3)))
}

fn remodel() -> Card {
    Card::new("Remodel", 4).with_type(CardType::Action(|game| {
        game.add_effect(Effect::Resolving {
            card: "Remodel".into(),
            request: ResolveRequest::ChooseHandCardsToDiscard {
                filter: CardFilter::Any,
            },
            player: ResolvingPlayer::ActivePlayer,
            effect: |game, message, _player, _request, _state| {
                if let ResolveReply::HandCardsChosen { choice } = message {
                    if choice.len() != 1 {
                        return NotResolved;
                    }

                    let card = game.players[game.active_player].hand.remove(choice[0]);
                    let cost = card.cost;

                    game.trash.push(card);

                    game.add_effect(Effect::Resolving {
                        card: "Remodel".into(),
                        request: ResolveRequest::GainCard {
                            filter: CardFilter::MaxCost(cost + 2),
                        },
                        player: ResolvingPlayer::ActivePlayer,
                        effect: |game, message, player, request, _state| {
                            if let ResolveReply::SupplyCardChosen { choice } = message {
                                if let ResolveRequest::GainCard { filter } = request {
                                    if let Ok(()) = game.gain_card(player, choice, filter) {
                                        return Resolved;
                                    }
                                }
                            }
                            NotResolved
                        },
                    });
                    return Resolved;
                }
                NotResolved
            },
        });
    }))
}

fn militia() -> Card {
    Card::new("Militia", 4)
        .with_type(CardType::Attack)
        .with_type(CardType::Action(|game| {
            coin!(game, 2);

            game.add_effect(Effect::Resolving {
                card: "Militia".into(),
                request: ResolveRequest::ChooseHandCardsToDiscard {
                    filter: CardFilter::Any,
                },
                player: ResolvingPlayer::AllNonActivePlayers,
                effect: |game, message, player, _request, _state| {
                    if let ResolveReply::HandCardsChosen { choice } = message {
                        if choice.len() == 0 {
                            if !game.players[player].hand.iter().any(|c| c.name == "Moat") {
                                return Resolved;
                            }
                        } else if game.players[player].hand.len() > 3
                            && choice.len() != game.players[player].hand.len() - 3
                        {
                            return NotResolved;
                        }

                        for c in choice.iter().sorted_by(|a, b| b.cmp(a)) {
                            game.players[player].discard(*c);
                        }

                        return Resolved;
                    }
                    NotResolved
                },
            });
        }))
}

fn market() -> Card {
    Card::new("Market", 5).with_type(CardType::Action(|game| {
        draw!(game, 1);
        action!(game, 1);
        buy!(game, 1);
        coin!(game, 1);
    }))
}

fn mine() -> Card {
    Card::new("Mine", 5).with_type(CardType::Action(|game| {
        game.add_effect(Effect::Resolving {
            card: "Mine".into(),
            request: ResolveRequest::ChooseHandCardsToDiscard {
                filter: CardFilter::Type(CardType::Treasure(0)),
            },
            player: ResolvingPlayer::ActivePlayer,
            effect: |game, message, _player, _request, state| {
                if let ResolveReply::HandCardsChosen { choice } = message {
                    if choice.len() != 1 {
                        return NotResolved;
                    }

                    if let Some(card) = game.get_active_player().hand.get(choice[0]) {
                        if let None = card.treasure() {
                            return NotResolved;
                        }
                    }

                    let card = game.players[game.active_player].hand.remove(choice[0]);
                    let cost = card.cost;

                    game.trash.push(card);

                    state.resolved = true;
                    game.add_effect(Effect::Resolving {
                        card: "Mine".into(),
                        request: ResolveRequest::GainCard {
                            filter: CardFilter::And(vec![
                                CardFilter::Type(CardType::Treasure(0)),
                                CardFilter::MaxCost(cost + 3),
                            ]),
                        },
                        player: ResolvingPlayer::ActivePlayer,
                        effect: |game, message, player, request, _state| {
                            if let ResolveReply::SupplyCardChosen { choice } = message {
                                if let ResolveRequest::GainCard { filter } = request {
                                    if let Ok(()) = game.gain_card(player, choice, filter) {
                                        return Resolved;
                                    }
                                }
                            }
                            NotResolved
                        },
                    });
                }
                NotResolved
            },
        });
    }))
}
